function reset_reactor(reactor) -- pre 2.0.11 compatible
	local state = reactor.state
	local temperature = reactor.core.temperature
	local surface = reactor.furnace.surface
	local name = reactor.furnace.name
	local position = reactor.furnace.position
	local force = reactor.furnace.force
	local quality = reactor.furnace.quality
	local input = reactor.furnace.get_inventory(defines.inventory.furnace_source)
	local output = reactor.furnace.get_output_inventory()
	local input1 = nil
	if input[1] and input[1].valid_for_read then
		input1 = {name = input[1].name, count = input[1].count, quality = input[1].quality}
	end
	local output1 = nil
	local output2 = nil
	if output[1] and output[1].valid_for_read then
		output1 = {name = output[1].name, count = output[1].count, quality = output[1].quality}
	end
	if output[2] and output[2].valid_for_read then
		output2 = {name = output[2].name, count = output[2].count, quality = output[2].quality}
	end
	----------------------------------
	local index = nil
	for i, r in pairs(storage.reactors) do
		if r.furnace == reactor.furnace then
			index = i
			break
		end
	end
	reactor.eccs_L.destroy() -- remove eccs
	reactor.eccs_R.destroy() -- remove eccs
	reactor.power.destroy() -- remove power entity
	if reactor.displayer.valid then reactor.displayer.destroy() end
	if reactor.core.valid then reactor.core.destroy() end
	if reactor.lamp.valid then reactor.lamp.destroy() end
	if reactor.light.valid then reactor.light.destroy() end
	if reactor.interface_warning and reactor.interface_warning.valid then reactor.interface_warning.destroy() end
	if reactor.cooling_warning and reactor.cooling_warning.valid then reactor.cooling_warning.destroy() end
	reactor.furnace.destroy() --only destroy when there should not remain a ghost
	local interface = reactor.interface
	storage.reactors[index] = nil
	local furnace = surface.create_entity{name = name, position = position, force = force, quality = quality}
	furnace.active = false
	add_reactor(furnace,interface)
	reactor = storage.reactors[interface.unit_number]
	reactor.core.temperature = temperature
	if input1 then
		furnace.get_inventory(defines.inventory.furnace_source).insert(input1)
	end
	if output1 then
		furnace.get_output_inventory().insert(output1)
	end
	if output2 then
		furnace.get_output_inventory().insert(output2)
	end
	change_reactor_state(state, reactor, {tick = game.tick})
	
end	





function to_burner(entity)
	local recipe, quality =entity.get_recipe()
	local fuel_cell_prototype = nil
	if recipe then
		local fuel_cell_name = recipe.ingredients[1].name
		fuel_cell_prototype = prototypes.item[fuel_cell_name]
		local fuel_value = fuel_cell_prototype.fuel_value
		
		return {
			heat = 1,
			fuel_value = fuel_value,
			remaining_burning_fuel = fuel_value * (1-entity.crafting_progress),
			currently_burning = {
				name = fuel_cell_prototype,
				quality = quality,
			}
		}
		
	else
		return {
			heat = 0,
			remaining_burning_fuel = 0,
			currently_burning = nil,
		}
	end
end

function respawn_furnace(reactor)
	local surface = reactor.furnace.surface
	local name = reactor.furnace.name
	local position = reactor.furnace.position
	local force = reactor.furnace.force
	local quality = reactor.furnace.quality
	local input = reactor.furnace.get_inventory(defines.inventory.furnace_source)
	local output = reactor.furnace.get_output_inventory()
	local input1 = nil
	if input[1] and input[1].valid_for_read then
		input1 = {name = input[1].name, count = input[1].count, quality = input[1].quality}
	end
	local output1 = nil
	local output2 = nil
	if output[1] and output[1].valid_for_read then
		output1 = {name = output[1].name, count = output[1].count, quality = output[1].quality}
	end
	if output[2] and output[2].valid_for_read then
		output2 = {name = output[2].name, count = output[2].count, quality = output[2].quality}
	end
	reactor.furnace.destroy()
	reactor.furnace = surface.create_entity{name = name, position = position, force = force, quality = quality}
	reactor.furnace.active = false
	if input1 then
		reactor.furnace.get_inventory(defines.inventory.furnace_source).insert(input1)
	end
	if output1 then
		reactor.furnace.get_output_inventory().insert(output1)
	end
	if output2 then
		reactor.furnace.get_output_inventory().insert(output2)
	end
end

function dbg (str)
	if not storage.dbg then storage.dbg = 1 end
		if type(str) ~= "number" and type(str) ~= "string" then
		if str == true then 
			str = "true" 
		elseif str == false then 
			str = "false" 
		else
			str = type(str)
		end
	end
	game.players[1].print(storage.dbg.."/"..game.tick..": "..str)
	storage.dbg = storage.dbg+1
end

function msg(s)
	game.print(s)
	--for _, player in pairs(game.players) do
	--	if player.connected then
	--		player.print(s)
	--	end
	--end
end

-- function logging(message)
	-- game.write_file("RealisticReactors.log","\r\n[" .. game.tick .. "] " .. message,true)
-- end

function deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

function compare_tables(o1, o2, ignore_mt)
    if o1 == o2 then return true end
    local o1Type = type(o1)
    local o2Type = type(o2)
    if o1Type ~= o2Type then return false end
    if o1Type ~= 'table' then return false end

    if not ignore_mt then
        local mt1 = getmetatable(o1)
        if mt1 and mt1.__eq then
            --compare using built in method
            return o1 == o2
        end
    end

    local keySet = {}

    for key1, value1 in pairs(o1) do
        local value2 = o2[key1]
        if value2 == nil or compare_tables(value1, value2, ignore_mt) == false then
            return false
        end
        keySet[key1] = true
    end

    for key2, _ in pairs(o2) do
        if not keySet[key2] then return false end
    end
    return true
end

function table_length(t)
	local i=0
	for _ in pairs(t) do
		i=i+1
	end
	return i
end

function find_compound_entity(entity, field)
	for rid, reactor in pairs(storage.reactors) do
		if reactor[field] == entity then
			return rid
		end
	end
end